/* Autor: Rafael Alberto Moreno Parra
   Fecha: 30 de Octubre de 2000

MOTOR18. Escrito en C++ (Compilado en Visual C++ 6.0)
Ambientes y Organismos (simples expresiones, algoritmos y mutacin)

Aqui se simula el comportamiento cclico de un ambiente.
Por ejemplo:
		
Ambiente = sin(y) * (4+3*X-X*X)
Genero un valor Y=PI/2 y doy valores de X = 1..10 entonces:
		
sin(PI/2) * (4+3*X-X*X) = 1 * (4+3*X-X*X) = 6,6,4,0,-6,-14,-24,-36,-50,-66
Ahora genero un organismo que se adapte a ese ambiente (6,6,4,0,-6,-14,-24,-36,-50,-66)

Luego cambio el valor de Y=PI/3 y doy valores de X=1..10 entonces obtengo el ambiente
5.1961, 5.1961, 3.4641, 0, -5.1961, -12.1243, -20.7846, -31.1769, -43.30127, -50.1576
Ahora genero otro organismo para dicho ambiente

Y asi sucesivamente de Y=0 hasta 2*PI. Con eso obtengo una serie de organismos que se adaptan
a un determinado ciclo de un ambiente cclico. Ahora es hacer que logren hacer simbiosis
y se obtiene el macroorganismo.
		
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "Organismo.h"
#include "MasApto.h"
#include "MedioAmbiente.h"
#include "Inicializa.h"

void Simular(void);
void vQuitaEspacios(char *sCadena);
void vLeft(char *cad1, char *cad2, char num_letras);
void vRight(char *cad1, char *cad2, char num_letras);
void vMid(char *cad1, char *cad2, char desde, char hasta);

#define PI 3.1415926537


//Programa Principal
int main()
{
	//Definicin de variables
	unsigned int iOrgSelecci=0; //Decide con que organismo jugar
	unsigned int iGeneracion=0; //Que estratgia usar para generar organismos
	unsigned int iNumGenes=0;  //Contador de genes(instrucciones a generar)
	unsigned int iMutables=0; //Nmero de instrucciones realmente ejecutadas
	unsigned int iCopia=0; //Para copiar de MasAptos a Organismo
	unsigned int iMutar=1; //Que gen mutar
	unsigned int iCicl=0; //En que parte del ciclo es mejor el organismo
	float fY=0; //Valor que tendr el ciclo en un momento
	float fAcumErr; //Acumula el error de aproximacin del organismo al ambiente
	int iX=0; //Valor X de la expresin de ambiente
	float fDiferen; //Diferencia entre ambiente y organismo
	float fMinimo; //Buscar la mejor adaptacin
	int iLugar; //Contador temporal para averiguar la mejor adpatacin
	float fAvance; //Nmero de partes en que se divide el ciclo
	unsigned int iRefrescaSemilla; //Para refrescar la semilla del generador aleatorio
	unsigned int iAcumRefresco;
	unsigned int iLleno;
	unsigned int iParcialLleno=0;
	unsigned int iFinSimul=1;
	float fTamanoCiclo; //Porcin del ciclo a evaluar, mximo de 0 a 2*PI


	Inicializa objInicializa; //Inicializa la simulacin
    MedioAmbiente objAmbiente;
	Organismo objOrganismo;

    //Presentacin
	objInicializa.vPantallaIni();
	
	//Lee los parametros de simulacin
	int iLeeDatos = objInicializa.vLeeArchivoIni();
	if (iLeeDatos == -1) return 1;
	objAmbiente.iInicio = objInicializa.stDatVA.iXini;
	objAmbiente.iFinal = objInicializa.stDatVA.iXfin;

	//Reserva memoria para el total de organismos
	MasApto objMasApto[100]; //Los mejores organismos
	
	//Evalua las partes en que se dividir el ciclo
	fTamanoCiclo = 2*PI/objInicializa.stDatVA.iParcCiclo;
	fAvance = fTamanoCiclo / objInicializa.stDatVA.iNumOrganismos;

	//Inicializa Organismos Aptos (cdigo -1)
	for (int iCont=0; iCont<objInicializa.stDatVA.iNumOrganismos+1; iCont++)
	{
		objMasApto[iCont].iCodMasApto = -1;
		objMasApto[iCont].fErrAdapta = (float) 99999999;
	}

    //Deduce el ambiente
	objAmbiente.vEvalAmbiente(objInicializa.stDatVA.iXini, objInicializa.stDatVA.iXfin, objInicializa.stDatVA.sAmbiente);


    //Inicializa el comportamiento aleatorio
	objOrganismo.IniciaSemillaT();
	objOrganismo.cMuta.vIniLista(	objInicializa.stDatVA.iPosibN,
									objInicializa.stDatVA.iPosibX,
									objInicializa.stDatVA.iPosibP);
	objOrganismo.m_iMaxiCiclos =  objInicializa.stDatVA.iNumCiclos;
	objOrganismo.vCreaADN(false, 65,
					objInicializa.stDatVA.iPosibIf,
					objInicializa.stDatVA.iPosibSet,
					objInicializa.stDatVA.iPosW,
					objInicializa.stDatVA.iPosX,
					objInicializa.stDatVA.iPosY,
					objInicializa.stDatVA.iPosZ,
					objInicializa.stDatVA.iPosIg,
					objInicializa.stDatVA.iPosMay,
					objInicializa.stDatVA.iPosMen,
					objInicializa.stDatVA.iPosDif,
					objInicializa.stDatVA.iLongExpr,
					objInicializa.stDatVA.iPosibX,
					objInicializa.stDatVA.iPosibP,
					objInicializa.stDatVA.iPosibN);

    //Informacin Inicial de la simulacin para simple expresin
	objOrganismo.m_oGen[1].bEjecuta = false;
	objOrganismo.m_oGen[1].cTipInst = 'S';
	objOrganismo.m_oGen[1].cVarActiva = 'X';
	objOrganismo.m_oGen[1].cVariable = 'Y';
	objOrganismo.m_oGen[1].iGotoLabel = 0;
	objOrganismo.m_iMaxGenOrg = 1;


	//Encabezado archivo de resultados
	objInicializa.vArchResult(); //Encabezado

	/* En esta simulacin ambiente cclico, se tendr en cuenta la generacin
	   de organismos simples (una lnea) y tipo algoritmo: Proceso de mutacin
	   fuerte y sutil */
	iRefrescaSemilla=0;
	iAcumRefresco=0;
	while(true)
	{
		//En pruebas muy largas, es necesario refrescar el generador de nmeros aleatorios
		iRefrescaSemilla++;
		if (iRefrescaSemilla > 1000)
		{
			iAcumRefresco++;
			iRefrescaSemilla = 0;
			printf("\nRefrescos (grupos de 1000): [%d]\n",iAcumRefresco);

			iFinSimul=1;
			for (iLleno=0; iLleno < objInicializa.stDatVA.iNumOrganismos; iLleno++)
			{
				printf("Asi esta: [%d] Error: [%f]\n", iLleno, objMasApto[iLleno].fErrAdapta);
				if (objMasApto[iLleno].fErrAdapta > objInicializa.stDatVA.fMaximoError)
					iFinSimul=0;
			}

			if (iAcumRefresco >= objInicializa.stDatVA.iDetiene) break;
			if (iFinSimul==1) break;

			objOrganismo.IniciaSemillaT(); //Inicializa la semilla por medio del reloj
		}

		//Se va a generar fijo: 2*PI / (PI/8) Organismos = 16 Organismos
		iOrgSelecci = rand()%objInicializa.stDatVA.iNumOrganismos; //Decide con que organismo jugar
        
		if (objMasApto[iOrgSelecci].iCodMasApto != -1) //Si ya lo ocupaba alguien
		{
			if (objMasApto[iOrgSelecci].m_iMaxGenOrg > 1 )
        		iGeneracion = rand()%5; //Decide entre las 5 estratgias
			else
      			iGeneracion = rand()%3; //Decide entre las 3 estratgias

			switch(iGeneracion)
			{
				case 0: //Genera un organismo
				    objOrganismo.cMuta.vCrearExpresion(objInicializa.stDatVA.iLongExpr, objInicializa.stDatVA.iPosibX, objInicializa.stDatVA.iPosibP, objInicializa.stDatVA.iPosibN);
				    strcpy(objOrganismo.m_oGen[1].sbExpresion,objOrganismo.cMuta.sExpresion);
					break;

				case 1: //Muta la simple expresion
					strcpy(objOrganismo.cMuta.sExpresion, objMasApto[iOrgSelecci].m_oGen[1].sbExpresion);
					objOrganismo.cMuta.vMutarPartes(objInicializa.stDatVA.iPosibX, objInicializa.stDatVA.iPosibN);
					strcpy(objOrganismo.m_oGen[1].sbExpresion,objOrganismo.cMuta.sExpresion);
					break;

				case 2: //Genera Aleatoriamente los Algoritmos
					iNumGenes = rand() % (objInicializa.stDatVA.iNumInstMax - objInicializa.stDatVA.iNumInstMin) + objInicializa.stDatVA.iNumInstMin;
					objOrganismo.vCreaADN(false, iNumGenes,
							objInicializa.stDatVA.iPosibIf,
							objInicializa.stDatVA.iPosibSet,
							objInicializa.stDatVA.iPosW,
							objInicializa.stDatVA.iPosX,
							objInicializa.stDatVA.iPosY,
							objInicializa.stDatVA.iPosZ,
							objInicializa.stDatVA.iPosIg,
							objInicializa.stDatVA.iPosMay,
							objInicializa.stDatVA.iPosMen,
							objInicializa.stDatVA.iPosDif,
							objInicializa.stDatVA.iLongExpr,
							objInicializa.stDatVA.iPosibX,
							objInicializa.stDatVA.iPosibP,
							objInicializa.stDatVA.iPosibN);

					//Restaura la primera instruccion
					objOrganismo.m_oGen[1].cOperacion = objMasApto[iOrgSelecci].m_oGen[1].cOperacion;
					objOrganismo.m_oGen[1].cTipInst = objMasApto[iOrgSelecci].m_oGen[1].cTipInst;
					objOrganismo.m_oGen[1].cVarActiva = objMasApto[iOrgSelecci].m_oGen[1].cVarActiva;
					objOrganismo.m_oGen[1].cVariable = objMasApto[iOrgSelecci].m_oGen[1].cVariable;
					objOrganismo.m_oGen[1].iGotoLabel = objMasApto[iOrgSelecci].m_oGen[1].iGotoLabel;
					strcpy(objOrganismo.m_oGen[1].sbExpresion,objMasApto[iOrgSelecci].m_oGen[1].sbExpresion);
					break;

				case 3: //Muta todo un Gen
					iMutables=0;
					for (iCopia=1; iCopia <= objMasApto[iOrgSelecci].m_iMaxGenOrg; iCopia++)
					{
						if (objMasApto[iOrgSelecci].m_oGen[iCopia].bEjecuta == true) iMutables++;
						objOrganismo.m_oGen[iCopia].bEjecuta = false;
						objOrganismo.m_oGen[iCopia].cOperacion = objMasApto[iOrgSelecci].m_oGen[iCopia].cOperacion;
						objOrganismo.m_oGen[iCopia].cTipInst = objMasApto[iOrgSelecci].m_oGen[iCopia].cTipInst;
						objOrganismo.m_oGen[iCopia].cVarActiva = objMasApto[iOrgSelecci].m_oGen[iCopia].cVarActiva;
						objOrganismo.m_oGen[iCopia].cVariable = objMasApto[iOrgSelecci].m_oGen[iCopia].cVariable;
						objOrganismo.m_oGen[iCopia].iGotoLabel = objMasApto[iOrgSelecci].m_oGen[iCopia].iGotoLabel;
						strcpy(objOrganismo.m_oGen[iCopia].sbExpresion,objMasApto[iOrgSelecci].m_oGen[iCopia].sbExpresion);
					}
					objOrganismo.m_iMaxGenOrg = objMasApto[iOrgSelecci].m_iMaxGenOrg;

					//Escoge que Gen mutar, muta los que se ejecutaron
					iMutar = rand() % (iMutables-1) + 2;
					for (iCopia=1; iCopia <= objMasApto[iOrgSelecci].m_iMaxGenOrg; iCopia++)
					{
						if (objMasApto[iOrgSelecci].m_oGen[iCopia].bEjecuta == true) iMutar--;
						if (iMutar==0) break;
					}
					iMutar = iCopia; //La posicin real a mutar

					//Muta el Gen
					objOrganismo.vHaceGen(iMutar, objOrganismo.m_iMaxGenOrg,
							objInicializa.stDatVA.iPosibIf,
							objInicializa.stDatVA.iPosibSet,
							objInicializa.stDatVA.iPosW,
							objInicializa.stDatVA.iPosX,
							objInicializa.stDatVA.iPosY,
							objInicializa.stDatVA.iPosZ,
							objInicializa.stDatVA.iPosIg,
							objInicializa.stDatVA.iPosMay,
							objInicializa.stDatVA.iPosMen,
							objInicializa.stDatVA.iPosDif,
							objInicializa.stDatVA.iLongExpr,
							objInicializa.stDatVA.iPosibX,
							objInicializa.stDatVA.iPosibP,
							objInicializa.stDatVA.iPosibN);
					break;

				case 4: //Muta un Gen pero de forma sutil
						iMutables=0;
						for (iCopia=1; iCopia <= objMasApto[iOrgSelecci].m_iMaxGenOrg; iCopia++)
						{
							if (objMasApto[iOrgSelecci].m_oGen[iCopia].bEjecuta == true) iMutables++;
							objOrganismo.m_oGen[iCopia].bEjecuta = false;
							objOrganismo.m_oGen[iCopia].cOperacion = objMasApto[iOrgSelecci].m_oGen[iCopia].cOperacion;
							objOrganismo.m_oGen[iCopia].cTipInst = objMasApto[iOrgSelecci].m_oGen[iCopia].cTipInst;
							objOrganismo.m_oGen[iCopia].cVarActiva = objMasApto[iOrgSelecci].m_oGen[iCopia].cVarActiva;
							objOrganismo.m_oGen[iCopia].cVariable = objMasApto[iOrgSelecci].m_oGen[iCopia].cVariable;
							objOrganismo.m_oGen[iCopia].iGotoLabel = objMasApto[iOrgSelecci].m_oGen[iCopia].iGotoLabel;
							strcpy(objOrganismo.m_oGen[iCopia].sbExpresion,objMasApto[iOrgSelecci].m_oGen[iCopia].sbExpresion);
						}
						objOrganismo.m_iMaxGenOrg = objMasApto[iOrgSelecci].m_iMaxGenOrg;

						//Escoge que Gen mutar, muta los que se ejecutaron
						iMutar = rand() % (iMutables - 1) + 2;
						for (iCopia=1; iCopia <= objMasApto[iOrgSelecci].m_iMaxGenOrg; iCopia++)
						{
							if (objMasApto[iOrgSelecci].m_oGen[iCopia].bEjecuta == true) iMutar--;
							if (iMutar==0) break;
						}
						iMutar = iCopia; //La posicin real a mutar

						//Muta el Gen en forma sutil
						objOrganismo.vMutaGen(iMutar, objOrganismo.m_iMaxGenOrg,
										objInicializa.stDatVA.iLongExpr,
										objInicializa.stDatVA.iPosibX,
										objInicializa.stDatVA.iPosibP,
										objInicializa.stDatVA.iPosibN);
						break;
			} //Fin switch()
		}
		else //Es la primera vez que se genera para esta parte del ciclo un organismo
		{
		    objOrganismo.cMuta.vCrearExpresion(objInicializa.stDatVA.iLongExpr, objInicializa.stDatVA.iPosibX, objInicializa.stDatVA.iPosibP, objInicializa.stDatVA.iPosibN);
		    strcpy(objOrganismo.m_oGen[1].sbExpresion,objOrganismo.cMuta.sExpresion);
		}

        //Evalua previamente el organismo
		objOrganismo.vEvaluaPrevio();

		// Busca en que parte del ciclo es mejor el organismo
		iCicl=0;
		fMinimo = 9999999;
		iLugar = 0;

		for (fY=0; fY<=fTamanoCiclo; fY+=fAvance)
		{
			fAcumErr=0;
			for (iX=objAmbiente.iInicio; iX<=objAmbiente.iFinal; iX++)
			{
				fDiferen = (float) fabs(objOrganismo.fEvalOrganismo((float)iX) - sin(fY)*objAmbiente.m_fAmbiente[iX]);
				fAcumErr += fDiferen;
			}

			// El organismo ocupa las posiciones en el ciclo donde mejor se desempee
			if (objMasApto[iLugar].fErrAdapta > fAcumErr)
			{
				for (iCopia=1; iCopia <= objOrganismo.m_iMaxGenOrg; iCopia++)
				{
					objMasApto[iLugar].m_oGen[iCopia].bEjecuta = objOrganismo.m_oGen[iCopia].bEjecuta;
					objMasApto[iLugar].m_oGen[iCopia].cOperacion = objOrganismo.m_oGen[iCopia].cOperacion;
					objMasApto[iLugar].m_oGen[iCopia].cTipInst = objOrganismo.m_oGen[iCopia].cTipInst;
					objMasApto[iLugar].m_oGen[iCopia].cVarActiva = objOrganismo.m_oGen[iCopia].cVarActiva;
					objMasApto[iLugar].m_oGen[iCopia].cVariable = objOrganismo.m_oGen[iCopia].cVariable;
					objMasApto[iLugar].m_oGen[iCopia].iGotoLabel = objOrganismo.m_oGen[iCopia].iGotoLabel;
					strcpy(objMasApto[iLugar].m_oGen[iCopia].sbExpresion, objOrganismo.m_oGen[iCopia].sbExpresion);
				}
				objMasApto[iLugar].fErrAdapta = fAcumErr;
				objMasApto[iLugar].iCodMasApto = iLugar;
				objMasApto[iLugar].m_iMaxGenOrg = objOrganismo.m_iMaxGenOrg;
				objOrganismo.IniciaSemillaR(); //Inicializa la semilla aleatoriamente
			}
			iLugar++;
		}

	}

	//Imprime los resultados
	char sTextOrganismo[5000];
	for (iLugar=0; iLugar < objInicializa.stDatVA.iNumOrganismos; iLugar++)
	{
		for (iCopia=1; iCopia <= objMasApto[iLugar].m_iMaxGenOrg; iCopia++)
		{
			objOrganismo.m_oGen[iCopia].cOperacion = objMasApto[iLugar].m_oGen[iCopia].cOperacion;
			objOrganismo.m_oGen[iCopia].cTipInst = objMasApto[iLugar].m_oGen[iCopia].cTipInst;
			objOrganismo.m_oGen[iCopia].cVarActiva = objMasApto[iLugar].m_oGen[iCopia].cVarActiva;
			objOrganismo.m_oGen[iCopia].cVariable = objMasApto[iLugar].m_oGen[iCopia].cVariable;
			objOrganismo.m_oGen[iCopia].iGotoLabel = objMasApto[iLugar].m_oGen[iCopia].iGotoLabel;
			strcpy(objOrganismo.m_oGen[iCopia].sbExpresion,objMasApto[iLugar].m_oGen[iCopia].sbExpresion);
		}
		objOrganismo.m_iMaxGenOrg = objMasApto[iLugar].m_iMaxGenOrg;
		objOrganismo.sDisplayADN(sTextOrganismo);
		objInicializa.vImprMacroOrg(iLugar, objInicializa.stDatVA.iNumOrganismos, sTextOrganismo, objMasApto[iLugar].fErrAdapta);
	}
	objInicializa.vImprTiempoFin();
	return 1;
}

